
function [T,P, Idx,Idx_children] = dendrogram_Ping(Z,TS,show,varargin)

    if(show == 1)
        subplot('Position',[0.05,.2,0.9,.75]);
        drawnow;
    end
    [T,P] = my_dendrogram(show,Z,varargin{:});
    
    % Idx(an Idx corresponds to a subtree)
    % 1. immediate child node1
    % 2. immediate child node2
    % 3. root node of this subtree
    % 4. total number of nodes below the root (# of nodes in this subtree
    %    excluding the root node)
    % 5. the height(distance) of the sub-subtree consisting of nodes Idx(i,1:3) (Z's last column, height)

    % Idx_children
    % 1. all child leaf nodes
    % 2. all child internal nodes
    % 3. the parent nodes of this subtree(the nodes to which this subtree is a child)
    
    n = size(Z,1); 
    len = length(TS);
    Idx = zeros(n,5);
    Idx_children = cell(n,3);
    k=n+1;
    XPos = T(:)';    
    Node = P(:)';       
    DistY = Z(:,end);
    DistX = zeros(size(DistY));
    
    for i=1:n
        k = k+1;       
        Idx(i,1) = Z(i,1);     % immediate child nodes
        Idx(i,2) = Z(i,2);
        Idx(i,3) = k;          % root node
        Idx(i,5) = Z(i,3);

        a = Idx(i,1);
        b = Idx(i,2);
        if(a<=len && b<=len)
            Idx(i,4) = 2;
            Idx_children{i,1} = [a;b];
            Idx_children{i,2} = [];
        elseif(a<=len && b> len)             
            [temp_idx,~] = find(Idx(:,3) == b);
            Idx(i,4) = 1+Idx(temp_idx,4);
            Idx_children{i,1} = [Idx_children{temp_idx,1}; a];
            Idx_children{i,2} = [Idx_children{temp_idx,2}, b];
        elseif(a>len && b<=len)
            [temp_idx,~] = find(Idx(:,3) == a);
            Idx(i,4) = 1+Idx(temp_idx,4);
            Idx_children{i,1} = [Idx_children{temp_idx,1}; b];
            Idx_children{i,2} = [Idx_children{temp_idx,2}, a];
        else 
            [temp_idx_a,~] = find(Idx(:,3) == a);
            [temp_idx_b,~] = find(Idx(:,3) == b);
            Idx(i,4) = Idx(temp_idx_a,4)+Idx(temp_idx_b,4);
            Idx_children{i,1} = [Idx_children{temp_idx_a,1}; Idx_children{temp_idx_b,1}];
            Idx_children{i,2} = [Idx_children{temp_idx_a,2}, Idx_children{temp_idx_b,2}, a, b];
        end    
        
        xa = XPos(find(Node==a,1));
        xb = XPos(find(Node==b,1));
        x_new = (xa+xb)/2;
        Node(k)=k;
        XPos(k)=x_new;        
        DistX(i) = x_new;
    end 
    
    % find this subtree's parent and ancester nodes
    for i = 1: n   
        parent = [];
        root = Idx(i,3);
        [a,~] = find(Idx(:,1:2) == root); % a is the index of the subtree to which current root is a child 
        while(~isempty(a))
            root = Idx(a,3);
            parent = [parent root];
            [a,~] = find(Idx(:,1:2) == root);
        end
        Idx_children{i,3} = parent;
    end
    
    % plot results  
    if(show == 1)
        sp_x = n/100;
        sp_y = max(Z(:,end))/50;
        for i=1:n
           %To Show Node ID
           str = sprintf('%d',i+n+1);
           text(DistX(i)+sp_x,DistY(i)+sp_y,str);       
        end
        axis off;
   
        n = length(TS);
        w = 0.9/(n+1);
        for i=1:n
            subplot('Position',[0.05+w*0.5+w*.05+w*(i-1),.01,w*0.9,.18]);
            plot(zscore(TS{P(i)}));title(P(i));
            ylim([-3 3]);
            axis off;
        end
    end

end