% magicScore: postive
% add textfile to record runtime information
% in dict, the trainData becomes index
% in dict, add the valid performance measure(non-overlap counting)
% add summarize same concept in Dict (count only non-overlap)
% add sum results over multiple runs
% add plot option
% add plot results option
% results print to screen or textfile

function [Dict,items,score_val,stats,n_denData] = mainFile_multiLabels_v4...
        (d,seqLen,dataFileName,labelFileName,W,MagicScore,Tfactor,...
        offlineLearn,showDen,showTopK,showResults,runNum,normalizeFunction)

    close all; clc;
    if(~exist('runNum','var'))
        runNum = 1;
    end
    if(~exist('normalizeFunction','var'))
        normalizeFunction = 'zscore';
    end
    dest = fullfile(d,['results_' num2str(runNum)]);
    if(~exist(dest,'dir'))
        mkdir(dest);
    end
    % load data 
    stream_data = importdata(fullfile(d,dataFileName)); 
    stream_labels = importdata(fullfile(d,labelFileName));
    
    % offline learn the mean-std and overlap distance threhold 
    if ( ((offlineLearn == 0) &  ~exist(fullfile(d,'mean-std.mat'),'file'))...
            |(offlineLearn == 1))
        NNDist = computeMeanStd(stream_data, seqLen, W, 6, 10000, 1);
        NNDist = [ [0 0]; NNDist];
        M = NNDist(:,1);  S = NNDist(:,2);
        save(fullfile(d,'mean-std.mat'),'M','S');
    else
        load(fullfile(d,'mean-std.mat'));
    end
    
    if(~exist(fullfile(d,'overlapTh.mat'),'file'))
        overlapTh = M(2);
        good = 0;
        while(~good)
            extractSeq(stream_data,seqLen,overlapTh);
            reply = input('Is this threshold good? Y/N [Y]: ', 's');
            if isempty(reply)
                reply = 'Y';
            end
            if(reply == 'y' | reply == 'Y')
                good = 1;
                save(fullfile(d,'overlapTh.mat'),'overlapTh');
            else  
                reply2 = '';
                while(isempty(reply2))
                    reply2 = input('Enter a new threhold: ', 's');
                    if(~isempty(reply2))
                        overlapTh = str2num(reply2);
                    end
                end
            end
        end     
    else
        load(fullfile(d,'overlapTh.mat'));
    end
    
    %Run the online algorithm to learn concepts
    textfile = fullfile(dest,sprintf(....
            'runtime_buffersize%d_score%2.2f_Tfactor%2.1f_run%d.txt',...
            W,MagicScore,Tfactor,runNum));
    [Dict,items,score_val,trueLabels,dataPos,n_denData] = myPatternFinder_multiLabels_v4(...
            stream_data,stream_labels,seqLen,overlapTh,W,MagicScore,...
            Tfactor,M,S,6,textfile,showDen,showTopK,normalizeFunction);
    
    % compute the statistics for each concept in Dict
    for i = 1:length(Dict)
        disp(Dict{i}.label)
        Dict{i}.pos_percent = 100 * Dict{i}.pos(1)/(length(stream_data)-W);        
        % raw statistics         
        totalIdx = find(trueLabels(Dict{i}.pos(2)+1:end,1)== Dict{i}.label);
        totalIdx = dataPos(totalIdx + Dict{i}.pos(2));               
        correctIdx = Dict{i}.dataIdx(find(Dict{i}.correct == 1));
        errorIdx = Dict{i}.dataIdx(find(Dict{i}.correct == 0));
        Dict{i}.totalNum = length(totalIdx);
        Dict{i}.tp = length(correctIdx);
        Dict{i}.fp = length(errorIdx);
        % valid statistics
        [validTotalNum,validTpIdx,validFpIdx] = countValidStatistics(...
                stream_labels,Dict{i},seqLen);
        Dict{i}.validCorrectIdx = validTpIdx;
        Dict{i}.validErrorIdx = validFpIdx;
        Dict{i}.validTotalNum = validTotalNum;
        Dict{i}.validTp = length(validTpIdx);
        Dict{i}.validFp = length(validFpIdx);
    end
    % summarize statistics 
    stats = containers.Map;
    for i = 1:length(Dict)
        clear perf;
        if(~isequal(Dict{i}.label,-1))  
            if(~isKey(stats,Dict{i}.label)) % first detection
                perf.label = Dict{i}.label;
                perf.totalNum = Dict{i}.validTotalNum;
                perf.pos = Dict{i}.pos(1);
                perf.pos_percent = Dict{i}.pos_percent;
                perf.dataIdx = Dict{i}.dataIdx;
                perf.dataDis = Dict{i}.dataDis;
                perf.correct = Dict{i}.correct;
                perf.variations = 1;
                if(length(stats) == 0)
                    stats = containers.Map(Dict{i}.label,perf);
                else
                    stats(Dict{i}.label) = perf;
                end
            else              
                perf = stats(Dict{i}.label);
                perf.dataIdx = [perf.dataIdx,Dict{i}.dataIdx];
                perf.dataDis = [perf.dataDis,Dict{i}.dataDis];
                perf.correct = [perf.correct,Dict{i}.correct];
                perf.variations = perf.variations + 1;
                stats(Dict{i}.label) = perf;
            end
        else  % inconsistents
            if(~isKey(stats,-1))
                perf.detectedTimes = 1;
                perf.dataIdx = Dict{i}.dataIdx;            
                perf.pos_percent = Dict{i}.pos_percent;
                if(length(stats) == 0)
                    stats = containers.Map(Dict{i}.label,perf);
                else
                    stats(Dict{i}.label) = perf;
                end
            else
                perf = stats(-1);
                perf.detectedTimes = perf.detectedTimes + 1;
                perf.dataIdx = [perf.dataIdx Dict{i}.dataIdx];
                stats(-1) = perf;
            end
        end
    end
    % compute valid summary statistics    
    labels = keys(stats);
    for i = 1:length(stats)     
        perf = stats(labels{i});    
        if(labels{i} ~= -1)
            [~,validTpIdx,validFpIdx] = countValidStatistics(stream_labels,perf,seqLen);
            perf.validCorrectIdx = validTpIdx;
            perf.validErrorIdx = validFpIdx;
            perf.tp = length(perf.validCorrectIdx);
            perf.fp = length(perf.validErrorIdx);
        else
            perf.validDataIdx = getNonOverlap(perf.dataIdx,seqLen);
            perf.nValidData = length(perf.validDataIdx);
        end
        stats(labels{i}) = perf;
    end
    
    %save results 
    save(fullfile(dest,sprintf('Dict_size%d_score%2.2f_Tfactor%d_run%d.mat',W,MagicScore,Tfactor,runNum)),'Dict');
    save(fullfile(dest,sprintf('stats_size%d_score%2.2f_Tfactor%d_run%d.mat',W,MagicScore,Tfactor,runNum)),'stats');
    save(fullfile(dest,sprintf('dataPos_size%d_score%2.2f_Tfactor%d_run%d.mat',W,MagicScore,Tfactor,runNum)),'dataPos');
    
    
    if(showResults == 1)
%         plotResults_Dict_v1(Dict,stream_data,seqLen);  
%         plotResult_stats(stats,stream_data,seqLen);
%         showResults_Dict(Dict);
%         showResults_stats(stats,seqLen);
    end
end


