% dictionary: add distance (each detection to the template)


function [Dict,items,score_val,trueLabels,dataPos,n_denData] = ...
    myPatternFinder_multiLabels_v4(stream_data,stream_labels,seqLen,...
    overlapTh,W,MagicScore,Tfactor,M,S,K,textfile,showDen,showTopK,normalizeFunction)

    if(~exist('normalizeFunction','var'))
        normalizeFunction = 'zscore';
    end
    normalize = str2func(normalizeFunction);
    
    % initialize
    fid = fopen(textfile,'a');
    overflow_num = 6; 
    stream_input = stream_data;    
    total_num = length(stream_input);
 
    Dict = [];
    items = cell(total_num-W,K);
    score_val = cell(1,total_num-W);
    trueLabels = [];
    dataPos = [];
    n_denData = 0;
    
    counter_increment = []; % counter_increment(1:seqLen) = template; counter_increment(end) = increment; 
    buffer = zeros(W, seqLen);
    bufferPos = zeros(W,1);
    
    % fill in the buffer and compute the linkage distance Y 
    currentPos = 1;
    currentTS = stream_input(1:seqLen);
    trueLabels = [trueLabels; stream_labels(1,:)];
    dataPos = [dataPos; 1];
    bufferPos(1) = 1;
    n_denData = n_denData + 1;   
    for i = 2: W
        [currentTS,currentPos] = addNextTS(stream_input,currentTS,currentPos,overlapTh);
        if(currentPos < total_num)
            buffer(i,:) = normalize(currentTS);   
            bufferPos(i) = currentPos;
            trueLabels = [trueLabels; stream_labels(currentPos,:)];  
            dataPos = [dataPos; currentPos];
            n_denData = n_denData + 1;
        end
    end
    Y = im_pdist_EU(buffer,W);     
    
    % ================================================== %
    %        iterate to process data one by one          %
    % ================================================== %
    while (currentPos < total_num)       
        % build dendragram; sort the subtrees; and find the valid ones
        if(showDen == 1)   
            figure(1);drawnow;
        end
        [Idx,Idx_children,sortedClusters] = denScore_Ping(buffer,Y,M,S,W,overflow_num,showDen); 
        
        topK_cand = validNode_v2(sortedClusters,Idx_children,Idx);   % find all valid subtrees, sorted in descending order
        no_plot = zeros(1,K);
        if length(topK_cand) < K
            diff = K-length(topK_cand);
            no_plot(K-diff+1:end) = 1;      
            topK_cand = [topK_cand zeros(1,K-length(topK_cand))];    
        end
        
        % update counter_increment
        counter_increment = update_counter(Idx,Idx_children,topK_cand(topK_cand~=0),...
                        K,overflow_num,buffer,counter_increment);
             
        % report the topK
        if(showTopK == 1 & showDen == 1)    
            figure(2);
        end
        protected_nodes = [];
        for j = 1: K 
            [items{i-W+1,j},score_val{i-W+1}(j)]=nell_report_2(showTopK,...
                    K,sortedClusters,Idx_children,buffer,seqLen,...
                    topK_cand(j),no_plot(j),j,counter_increment); 
            protected_nodes = [protected_nodes; items{i-W+1,j}];
        end   
        if(showTopK == 1)
            drawnow;
        end
        disp(['top-1: ' num2str(score_val{i-W+1}(1))]);    
              
        % ============================================== %
        % if Query is triggered by Top_1
        if(score_val{i-W+1}(1) >= MagicScore)          
            % query user
            query_data = buffer(items{i-W+1,1},:);
            labels = query_labels_multiLabels(bufferPos(items{i-W+1,1}),stream_labels);
            % form a classifier
            template = mean(query_data);
            nItems = size(items{i-W+1,1});
            T = score_val{i-W+1}(1) * S(nItems) + M(nItems);          
            % update dictionary
            n = length(Dict);
            Dict{n+1}.template = template;
            Dict{n+1}.threshold = Tfactor *T(1);        
            Dict{n+1}.pos = [currentPos,i];
            Dict{n+1}.trainDataIdx = bufferPos(items{i-W+1,1});
            Dict{n+1}.dataIdx = [];
            Dict{n+1}.correct = [];
            Dict{n+1}.dataDis = [];
            [flag_consistent commonlabel] = isconsistent(labels); 
            if(flag_consistent)  
                % consistent pattern
                Dict{n+1}.label = commonlabel(1);
                for kk = 2:length(commonlabel)
                    Dict{n+kk} = Dict{n+1};
                    Dict{n+kk}.label = commonlabel(kk);
                end
                disp([num2str(i-W+1) ' remove a pattern...']);            
            else
                % inconsitent pattern
                Dict{n+1}.label = -1;
                Dict{n+1}.userlabels = labels;
                disp([num2str(i-W+1) ' remove an inconsistency...']);
            end
     
            % remove the Top_1 items and refill the buffer             
            C = items{i-W+1,1};     % buffer position to be replaced
            kk = 1;
            while(kk <= length(C))
                cand = C(kk); 
                if(length(Dict)>0)
                    startPos = max(max(Dict{length(Dict)}.trainDataIdx)+seqLen-1,currentPos);
                else
                    startPos = currentPos;
                end
                [currentTS,currentPos] = addNextTS(stream_input,currentTS,...
                    startPos,overlapTh);  
                if(currentPos <= total_num) 
                    i = i + 1;           
                    trueLabels = [trueLabels; stream_labels(currentPos,:)];  
                    dataPos = [dataPos;currentPos];
                    new_ts = normalize(currentTS);
                    [flag_in,Dict,correct] = preprocess_multiLabels_v4(Dict,...
                        new_ts,currentPos,stream_labels(currentPos,:));
                    if(flag_in ~= 1)     % classify to a concept in dictionary
                        if(correct == 1)
                            temp = 'correctly classified';
                        else
                            temp = 'misclassified';
                        end
                        disp(['data ' num2str(i) ' is ' temp]);
                    else
                        buffer(cand,:) = new_ts;
                        bufferPos(cand) = currentPos;
                        n_denData = n_denData + 1;
                        kk = kk + 1;
                        disp([num2str(i-W+1) ' remove: ' num2str(cand)]); 
                        fprintf(fid,['\r\n' num2str(i-W+1) ' remove: ' num2str(cand)]);
                    end
                else    
                    fclose(fid);
                    return;
                end
            end
               
            % rebuild dendrogram
            Y = im_pdist_EU(buffer,W);
                 
        % ============================================== %   
        % if no query is triggered
        else
            flag_in = 0;
            while(flag_in ~= 1)
                % extract the next data
                if(length(Dict)>0)
                    startPos = max(max(Dict{length(Dict)}.trainDataIdx)+seqLen-1,currentPos);
                else
                    startPos = currentPos;
                end
                [currentTS,currentPos] = addNextTS(stream_input,currentTS,...
                    startPos,overlapTh);  
                 if(currentPos > total_num)
                     fclose(fid);
                     return;
                 end
                 i = i + 1;           
                 trueLabels = [trueLabels; stream_labels(currentPos,:)]; 
                 dataPos = [dataPos; currentPos];
                 new_ts = normalize(currentTS);

                % compare to the dictionary
                [flag_in,Dict,correct] = preprocess_multiLabels_v4(Dict,new_ts,...
                    currentPos,stream_labels(currentPos,:));
                % if classify to the dictionary
                if(flag_in ~= 1) 
                    if(correct == 1)
                        temp = 'correctly classified';
                    else
                        temp = 'misclassified';
                    end
                    disp(['data ' num2str(i) ' is ' temp]);

                % if throw into dendrogram
                else
                    D = pdist2(buffer,normalize(new_ts));
                    [~,NN_idx] = min(D);
                    if(~ismember(NN_idx,protected_nodes)) % not close to topK
                        overflow_flag = 0;
                    else
                        [overflow_flag,subtree_idx] = isoverflow(NN_idx,...
                            Idx_children,Idx,topK_cand,K,overflow_num);
                    end

                    % overflow, keep the best subset
                    if(overflow_flag == 1)
                        leastTightNode = findLeastTightNode(buffer,...
                                Idx_children{subtree_idx,1},normalize(new_ts));
                        if(leastTightNode == -1)
                            % discard the new data, and increase the counter
                            rm_cand = -1;
                            template = mean(buffer(Idx_children{subtree_idx,1}));
                            for j = 1:size(counter_increment)
                                if(computeED(counter_increment(j,1:end-1),template) == 0)
                                    counter_increment(j,end) = counter_increment(j,end) + 1;
                                end
                            end     
                            disp([num2str(i-W+1) ' overflow: no data is replaced; subtree counter increased']);
                            fprintf(fid,['\r\n' num2str(i-W+1) 'overflow: no data is replaced; subtree counter increased']);  
                        else
                            % otherwise, replace the leastTightNode in the buffer
                            rm_cand = leastTightNode;
                            buffer(leastTightNode,:) = normalize(new_ts);
                            bufferPos(leastTightNode) = currentPos;
                            disp([num2str(i-W+1) ' overflow: remove ' num2str(leastTightNode)]);
                            fprintf(fid,['\r\n' num2str(i-W+1) ' overflow: remove ' num2str(leastTightNode)]);
                        end

                    % not overflow, randomly throw one-protected or low-rank data           
                    else   
                        if(length(protected_nodes) < W)   % randomly remove one un-protected data
                            C = setdiff(1:W,protected_nodes);  
                        else                              % remove one from low-rank subtree
                            valid_topK = topK_cand(topK_cand~=0);
                            low_rank = valid_topK(end);
                            [temp_ix,~] = find(Idx(:,3) == low_rank);
                            C = Idx_children{temp_ix,1};
                        end
                        num_rm = length(C);
                        rm_idx = randi(num_rm);
                        rm_cand = C(rm_idx);
                        buffer(rm_cand,:) = normalize(new_ts);
                        bufferPos(rm_cand) = currentPos;      
                        disp([num2str(i-W+1) ' remove: ' num2str(rm_cand)]); 
                        fprintf(fid,['\r\n' num2str(i-W+1) ' remove: ' num2str(rm_cand)]);
                    end
                    n_denData = n_denData + 1;

                    % update the linkage distance
                    if(rm_cand ~= -1)
                        Y = im_pdist_EU_update(Y,D,rm_cand);
                    end
                end
            end
        end
    end
    fclose(fid);
end

